from memory_profiler import memory_usage, profile
from sqlalchemy import create_engine

from pynYNAB.Client import clientfromargs
from pynYNAB.schema.budget import Transaction


class Args(object):
    email='email'
    password='password'
    nynabconnection=None
    budgetname='test budget'

client = clientfromargs(Args(),sync=False)
#client.session.bind.echo = True

@profile
def func():
    print('test_sync')
    for i in range(0,5000):
        client.budget.be_transactions.append(Transaction())
    client.session.commit()
    client.budget.clear_changed_entities()
    print('Entities in the client: %i'%(sum([len(getattr(client.budget, f)) for f in client.budget.listfields]) +
          sum([len(getattr(client.catalog, f)) for f in client.catalog.listfields])))
    pass

if __name__ == '__main__':
    func()







import json
import logging
from datetime import datetime

import re

from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.sql.sqltypes import Enum as sqlaEnum, String, DateTime
from aenum import Enum
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import event
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm.base import ONETOMANY, MANYTOMANY

from pynYNAB import KeyGenerator
from pynYNAB.schema.types import AmountType
from sqlalchemy import inspect

LOG = logging.getLogger(__name__)

class AccountTypes(Enum):
    undef = 'undef'
    Checking = 'Checking'
    Savings = 'Savings'
    CreditCard = 'CreditCard'
    Cash = 'Cash'
    LineOfCredit = 'LineOfCredit'
    PayPal = 'PayPal'
    MerchantAccount = 'MerchantAccount'
    InvestmentAccount = 'InvestmentAccount'
    Mortgage = 'Mortgage'
    OtherAsset = 'OtherAsset'
    OtherLiability = 'OtherLiability'


on_budget_dict = dict(
    undef=None,
    Checking=True,
    Savings=True,
    CreditCard=True,
    Cash=True,
    LineOfCredit=True,
    Paypal=True,
    MerchantAccount=True,
    InvestmentAccount=False,
    Mortgage=False,
    OtherAsset=False,
    OtherLiability=False,
)
on_budget_dict[None] = None


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Entity):
            return obj.get_apidict()
        elif isinstance(obj, Enum):
            return obj.value
        else:
            return json.JSONEncoder.default(self, obj)


class UnknowEntityFieldValueError(Exception):
    pass


class BaseModel(object):
    id = Column(String, primary_key=True, default=KeyGenerator.generateuuid)
    is_tombstone = Column(Boolean, default=False)

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    @property
    def listfields(self):
        relations = inspect(self.__class__).relationships
        return {k: relations[k].mapper.class_ for k in relations.keys() if
                relations[k].direction == ONETOMANY or relations[k].direction == MANYTOMANY}

    @property
    def scalarfields(self):
        scalarcolumns = self.__table__.columns
        return {k: scalarcolumns[k].type.__class__ for k in scalarcolumns.keys() if k != 'parent_id' and k != 'knowledge_id'}



def configure_listener(mapper, class_):
    for col_attr in mapper.column_attrs:
        column = col_attr.columns[0]
        if column.default is not None:
            default_listener(col_attr, column.default)

    for rel_attr in mapper.relationships:
        expectedtype_listener(rel_attr)


def expectedtype_listener(rel_attr):
    @event.listens_for(rel_attr, 'append')
    def append(target, value, initiator):
        expected_type = initiator.parent_token.mapper.class_
        value_type = type(value)
        if expected_type != value_type:
            raise ValueError('type %s, attribute %s, expect a %s, received a %s ' % (
            type(target), rel_attr.key, expected_type, value_type))


def default_listener(col_attr, default):
    """Establish a default-setting listener.

    Given a class_, attrname, and a :class:`.DefaultGenerator` instance.
    The default generator should be a :class:`.ColumnDefault` object with a
    plain Python value or callable default; otherwise, the appropriate behavior
    for SQL functions and defaults should be determined here by the
    user integrating this feature.

    """

    @event.listens_for(col_attr, "init_scalar", retval=True, propagate=True)
    def init_scalar(target, value, dict_):

        if default.is_callable:
            # the callable of ColumnDefault always accepts a context
            # argument; we can pass it as None here.
            value = default.arg(None)
        elif default.is_scalar:
            value = default.arg
        else:
            # default is a Sequence, a SQL expression, server
            # side default generator, or other non-Python-evaluable
            # object.  The feature here can't easily support this.   This
            # can be made to return None, rather than raising,
            # or can procure a connection from an Engine
            # or Session and actually run the SQL, if desired.
            raise NotImplementedError(
                "Can't invoke pre-default for a SQL-level column default")

        # set the value in the given dict_; this won't emit any further
        # attribute set events or create attribute "history", but the value
        # will be used in the INSERT statement
        dict_[col_attr.key] = value

        # return the value as well
        return value


re_uuid = re.compile('[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')
re_date = re.compile(r'\d{4}[\/ .-]\d{2}[\/.-]\d{2}')



def date_from_api(columntype, string):
    result = re_date.search(string)
    if result is not None:
        return datetime.strptime(result.group(0), '%Y-%m-%d').date()


fromapi_conversion_functions_table = {
    Date: date_from_api,
    DateTime: lambda t,x: datetime.strptime(x, '%Y-%m-%dT%H:%M:%S.%f'),
    AmountType: lambda t, x: float(x) / 1000,
    sqlaEnum: lambda t, x: t.enum_class[x]
}

toapi_conversion_functions_table = {
    Date: lambda t, x: x.strftime('%Y-%m-%d'),
    DateTime: lambda t, x: x.strftime('%Y-%m-%dT%H:%M:%S.%f'),
    AmountType: lambda t, x: int(float(x) * 1000),
    sqlaEnum: lambda t, x: x._name_
}


class Entity(BaseModel):
    def get_apidict(self):
        entityDict = self.get_dict()
        for column in self.__table__.columns:
            if column.name in entityDict and entityDict[column.name] is not None:
                conversion_function = toapi_conversion_functions_table.get(column.type.__class__, lambda t, x: x)
                entityDict[column.name] = conversion_function(column.type, entityDict[column.name])
        return entityDict

    def get_dict(self):
        return {key: getattr(self, key) for key in self.scalarfields if key != 'parent_id'}

    def __unicode__(self):
        return self.__str__()

    def __str__(self):
        return (type(self), self.id).__str__()

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        try:
            return self.key() == other.key()
        except:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def key(self,excludes=None):
        t = tuple()
        for k, v in self.get_dict().items():
            if excludes and k in excludes:
                continue
            if isinstance(v, list):
                t += tuple(v)
            else:
                t += (v,)
        return t

    @property
    def key2(self):
        return self.key(excludes=['id'])

    def __hash__(self):
        return hash(self.key())

    def copy(self):
        returnvalue = type(self)(**self.get_dict())
        return returnvalue

    @classmethod
    def from_dict(cls, entitydict):
        return cls(**entitydict)

    @classmethod
    def from_apidict(cls, entityDict):
        return cls.from_dict(cls.apidict_toinsertdict(entityDict))

    @classmethod
    def apidict_toinsertdict(cls, entityDict):
        modified_dict = {}
        for column in cls.__table__.columns:
            if column.name in entityDict and entityDict[column.name] is not None:
                conversion_function = fromapi_conversion_functions_table.get(column.type.__class__, lambda t, x: x)
                modified_dict[column.name] = conversion_function(column.type, entityDict[column.name])
        return modified_dict


Base = declarative_base(cls=Entity)

event.listen(BaseModel, 'mapper_configured', configure_listener, propagate=True)


class DictDiffer(object):
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    def __init__(self, current_dict, past_dict):
        self.current_dict, self.past_dict = current_dict, past_dict
        self.set_current, self.set_past = set(current_dict.keys()), set(past_dict.keys())
        self.intersect = self.set_current.intersection(self.set_past)

    def added(self):
        return self.set_current - self.intersect

    def removed(self):
        return self.set_past - self.intersect

    def changed(self):
        return set(o for o in self.intersect if self.past_dict[o] != self.current_dict[o])


class RootEntity(BaseModel):
    previous_map = {}

    def _get_changed(self, fn):
        changed_entities = self.get_changed_entities()
        changed_dict = {}
        for key in changed_entities:
            changed_dict[key] = list(map(fn, changed_entities[key]))
        return changed_dict

    def get_changed_apidict(self):
        return self._get_changed(lambda entity: entity.get_apidict())

    def get_changed_dict(self):
        return self._get_changed(lambda entity: entity.get_dict())

    def get_changed_entities(self):
        current_map = self.getmaps()
        diff_map = {}

        for key in current_map:
            if key not in diff_map:
                diff_map[key] = {}
            if isinstance(current_map[key], dict):
                if key in self.previous_map:
                    diff = DictDiffer(current_map[key], self.previous_map[key])
                    for obj_id in diff.added() | diff.changed():
                        obj = current_map[key][obj_id]
                        objc = obj.copy()
                        diff_map[key][obj_id] = objc
                    for obj_id in diff.removed():
                        obj = self.previous_map[key][obj_id]
                        objc = obj.copy()
                        objc.is_tombstone = True
                        diff_map[key][obj.id] = objc

                else:
                    diff_map[key] = current_map[key]
        returnvalue = {}
        for key, value in diff_map.items():
            if isinstance(value, dict):
                if value:
                    returnvalue[key] = list(value.values())

        return returnvalue

    def __init__(self):
        super(RootEntity, self).__init__()
        self.clear_changed_entities()

    @orm.reconstructor
    def clear_changed_entities(self):
        self.previous_map = self.getmaps()

    def getmap(self, key):
        objs_dict = {}
        if getattr(self, key) is not None:
            for instance in getattr(self, key):
                objc = instance.copy()
                objs_dict[str(instance.id)] = objc
        return objs_dict

    def getmaps(self):
        objs_dict = {}
        for key in self.listfields:
            objs_dict[key] = self.getmap(key)
        for key in self.scalarfields:
            objs_dict[key] = getattr(self, key)
        return objs_dict

    @declared_attr
    def knowledge_id(self):
        return Column(ForeignKey('knowledge.id'))

    @declared_attr
    def knowledge(self):
        return relationship('Knowledge')



import csv
import json
import os
import sys
from collections import namedtuple
from datetime import datetime

import logging
from jsontableschema.exceptions import InvalidSchemaError
from jsontableschema.model import SchemaModel

from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.schema.budget import Payee, Transaction

scriptsdir = os.path.dirname(os.path.abspath(__file__))
schemas_dir = os.path.join(scriptsdir, 'csv_schemas')

LOG = logging.getLogger(__name__)

def do_csvimport(args, client=None):
    delta = 0
    if client is None:
        client = clientfromargs(args)

    LOG.debug('selected schema %s' % (args.schema,))
    if os.path.exists(args.schema):
        schemafile = args.schema
    else:
        schemafile = os.path.join(schemas_dir, args.schema + '.json')
        if not os.path.exists(schemafile):
            LOG.error('This schema doesn''t exist in csv_schemas')
            exit(-1)
    try:
        schema = SchemaModel(schemafile, case_insensitive_headers=True)
        with open(schemafile, 'r') as sf:
            schemacontent = json.load(sf)
            try:
                nheaders = schemacontent['nheaders']
            except KeyError:
                nheaders = 1
    except InvalidSchemaError:
        LOG.error('Invalid CSV schema')
        raise
    LOG.debug('schema headers %s' % schema.headers)

    if 'account' not in schema.headers and args.accountname is None:
        LOG.error('This schema does not have an account column and no account name was provided')
        exit(-1)

    accounts = {x.account_name: x for x in client.budget.be_accounts}
    payees = {p.name: p for p in client.budget.be_payees}
    mastercategories_perid = {m.id: m for m in client.budget.be_master_categories}
    subcategories = {}
    for s in client.budget.be_subcategories:
        m = mastercategories_perid[s.entities_master_category_id]
        subcategories[m.name + ':' + s.name] = s

    def getaccount(accountname):
        try:
            LOG.debug('searching for account %s' % accountname)
            return accounts[accountname]
        except KeyError:
            LOG.error('Couldn''t find this account: %s' % accountname)
            exit(-1)

    def getpayee(payeename):
        try:
            LOG.debug('searching for payee %s' % payeename)
            return payees[payeename]
        except KeyError:
            LOG.debug('Couldn''t find this payee: %s' % payeename)
            payee = Payee(name=payeename)
            client.budget.be_payees.append(payee)
            delta += 1
            return payee

    def getsubcategory(categoryname):
        try:
            LOG.debug('searching for subcategory %s' % categoryname)
            return subcategories[categoryname]
        except KeyError:
            LOG.debug('Couldn''t find this category: %s' % categoryname)
            exit(-1)

    entities_account_id = None
    if 'account' not in schema.headers:
        entities_account_id = getaccount(args.accountname).id

    amount = None
    if 'inflow' in schema.headers and 'outflow' in schema.headers:
        pass
    elif 'amount' in schema.headers:
        pass
    else:
        LOG.error('This schema doesn''t provide an amount column or (inflow,outflow) columns')
        exit(-1)

    csvrow = namedtuple('CSVrow', field_names=schema.headers)

    imported_date = datetime.now().date()

    LOG.debug('OK starting the import from %s ' % os.path.abspath(args.csvfile))
    with open(args.csvfile, 'r') as inputfile:
        header = []
        for i in range(0, nheaders):
            header.append(inputfile.readline())
        for row in csv.reader(inputfile):
            if sys.version[0] == '2':
                row = [cell.decode('utf-8') for cell in row]
            if all(map(lambda x: x.strip() == '', row)):
                continue
            LOG.debug('read line %s' % row)
            result = csvrow(*list(schema.convert_row(*row, fail_fast=True)))
            if 'account' in schema.headers:
                entities_account_id = getaccount(result.account).id
            if entities_account_id is None:
                LOG.error(
                    'No account id, the account %s in the an account column was not recognized' % result.account)
                exit(-1)
            if 'inflow' in schema.headers and 'outflow' in schema.headers:
                amount = result.inflow - result.outflow
            elif 'amount' in schema.headers:
                amount = result.amount

            if 'category' in schema.headers and result.category:
                entities_subcategory_id = getsubcategory(result.category).id
            else:
                entities_subcategory_id = None
            if 'payee' in schema.headers:
                imported_payee = result.payee
            else:
                imported_payee = ''
            entities_payee_id = getpayee(imported_payee).id
            if 'memo' in schema.headers:
                memo = result.memo
            else:
                memo = ''

            transaction = Transaction(
                entities_account_id=entities_account_id,
                cash_amount=amount,
                amount=amount,
                date=result.date,
                entities_payee_id=entities_payee_id,
                entities_subcategory_id=entities_subcategory_id,
                imported_date=imported_date,
                imported_payee=imported_payee,
                memo=memo,
                source="Imported"
            )

            if args.import_duplicates or not (transaction.key2 in [tr.key2 for tr in client.budget.be_transactions]):
                LOG.debug('Appending transaction %s ' % transaction.get_dict())
                client.budget.be_transactions.append(transaction)
                delta += 1
            else:
                LOG.debug('Duplicate transaction found %s ' % transaction.get_dict())

    return delta


if __name__ == "__main__":
    from pynYNAB.__main__ import MainCommands
    MainCommands.csvimport()




import json
import uuid

from sqlalchemy import String, types
from sqlalchemy import TypeDecorator


class ArrayType(TypeDecorator):
    """ Sqlite-like does not support arrays.
        Let's use a custom type decorator.

        See http://docs.sqlalchemy.org/en/latest/core/types.html#sqlalchemy.types.TypeDecorator
    """
    impl = String

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        return json.loads(value)

    def copy(self, **kw):
        return ArrayType(self.impl.length)


class AmountType(types.TypeDecorator):
    impl = types.Integer

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(types.Integer)

    def process_bind_param(self, value, dialect):
        return int(value * 1000) if value is not None else None

    def process_result_value(self, value, dialect):
        return float(value) / 1000 if value is not None else None



import random
import time
from datetime import datetime

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from pynYNAB.schema.Entity import AccountTypes
from pynYNAB.schema.budget import Account, Payee


def rate_limited(maxpersecond):
    minInterval = 1.0 / float(maxpersecond)

    def decorate(func):
        lastTimeCalled = [None]

        def rateLimitedFunction(*args, **kargs):
            if lastTimeCalled[0] is not None:
                elapsed = time.clock() - lastTimeCalled[0]
                leftToWait = minInterval - elapsed
                if leftToWait > 0:
                    time.sleep(leftToWait)
            ret = func(*args, **kargs)
            lastTimeCalled[0] = time.clock()
            return ret

        return rateLimitedFunction

    return decorate


def get_or_create_account(client, name):
    accounts = {a.account_name: a for a in client.budget.be_accounts if
                a.account_name == name}
    if name in accounts:
        account = accounts[name]
        client.delete_account(account)

    account = Account(
        account_type=AccountTypes.Checking,
        account_name=name
    )

    client.add_account(account, balance=random.randint(-10, 10), balance_date=datetime.now())
    return account


def get_or_create_payee(client, name):
    payees = {p.name: p for p in client.budget.be_payees if
              p.name == name}
    if name in payees:
        return payees[name]
    payee = Payee(
        name=name
    )

    client.budget.be_payees.append(payee)
    client.push(1)
    return payee

def get_one_or_create(session,
                      model,
                      create_method='',
                      create_method_kwargs=None,
                      **kwargs):
    try:
        return session.query(model).filter_by(**kwargs).one(), False
    except NoResultFound:
        kwargs.update(create_method_kwargs or {})
        created = getattr(model, create_method, model)(**kwargs)
        try:
            session.add(created)
            session.flush()
            return created, True
        except IntegrityError:
            session.rollback()
            return session.query(model).filter_by(**kwargs).one(), True


import unittest

import re

from pynYNAB.connection import nYnabConnection, NYnabConnectionError
from mock import Mock

noerror = {'error':None}

new_post= Mock()

class MockResponse(object):
    elapsed=0
    text=''

    def __init__(self, returnvalue=noerror, status_code=200, headers={}):
        self.jsonreturn = returnvalue
        self.status_code = status_code
        self.headers = headers

    def json(self):
        return self.jsonreturn

connection = nYnabConnection(email='email', password='password')
opname = 'opname'
request_dict = {}
connection.session.post = new_post


class TestConnection(unittest.TestCase):
    def test_init_session(self):
        uuidhex = re.compile('[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}', re.I)
        user_id = '123456'
        session_token = 'abcdef'

        # mock response
        def new_do_request( request_dic, opname):
            self.assertEqual(opname,'loginUser')
            self.assertTrue(uuidhex.match(request_dic['device_info']['id']))
            del request_dic['device_info']
            self.assertEqual(request_dic,{'email': 'email',
                                          'password': 'password',
                                          'remember_me': True})
            return {'session_token':session_token,'user':{'id':user_id}}

        connection.dorequest = new_do_request
        connection.init_session()
        self.assertEqual(connection.user_id,user_id)
        self.assertEqual(connection.sessionToken, session_token)

    def test_dorequest(self):
        new_post.side_effect = [MockResponse()]
        returnvalue_request = connection.dorequest(request_dict,opname)
        self.assertEqual(returnvalue_request,noerror)

    def test_dorequest_fail(self):
        new_post.side_effect=[MockResponse(status_code=500)]
        self.assertRaises(NYnabConnectionError, lambda:connection.dorequest(request_dict,opname))

    def test_connect_error_no_id(self):
        new_post.side_effect=[MockResponse({'error':{'message':'meh'}})]
        self.assertRaises(NYnabConnectionError, lambda:connection.dorequest(request_dict,opname))

    def test_connect_error_user_password_invalid(self):
        new_post.side_effect=[MockResponse({'error':{'id':'user_password_invalid'}})]
        self.assertRaises(NYnabConnectionError, lambda:connection.dorequest(request_dict,opname))

    def test_connect_error_user_not_found(self):
        new_post.side_effect=[MockResponse({'error':{'id':'user_not_found'}})]
        self.assertRaises(NYnabConnectionError, lambda:connection.dorequest(request_dict,opname))



import unittest

from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.__main__ import parser
from pynYNAB.schema import DictDiffer
from pynYNAB.schema.budget import Transaction
from test_live.common import CommonLive
from test_live.common import needs_account


# noinspection PyArgumentList
class LiveTests(CommonLive):
    @needs_account()
    def test_add_deletetransaction(self):
        from datetime import datetime
        transaction = Transaction(
            amount=1,
            cleared='Uncleared',
            date=datetime.now(),
            entities_account_id=self.account.id,
        )
        self.client.add_transaction(transaction)
        self.reload()
        self.assertIn(transaction, self.client.budget.be_transactions)
        self.client.delete_transaction(transaction)
        self.reload()
        self.assertNotIn(transaction, self.client.budget.be_transactions)


class LiveTests2(unittest.TestCase):
    def test_roundtrip(self):
        args = parser.parse_known_args()[0]

        # 1. gets sync data from server
        # 2. tests that to_api(from_api(data)) is the same thing


        client = clientfromargs(args, sync=False)
        sync_data = client.catalogClient.get_sync_data_obj()
        budget_version_id = next(d['id'] for d in sync_data['changed_entities']['ce_budget_versions'] if
                                 d['version_name'] == args.budgetname)
        client.budget_version_id = budget_version_id

        for objclient in (client.budgetClient, client.catalogClient):
            sync_data = objclient.get_sync_data_obj()
            server_changed_entities = sync_data['changed_entities']

            for key in server_changed_entities:
                if key in objclient.obj.listfields:
                    if len(server_changed_entities[key]) == 0:
                        continue
                    obj_dict = server_changed_entities[key][0]
                    typ = objclient.obj.listfields[key]
                    obj_dict2 = typ.from_apidict(obj_dict).get_apidict()

                    diff = DictDiffer(obj_dict2, obj_dict)
                    for k in diff.changed():
                        AssertionError('changed {}: {}->{}'.format(k, obj_dict[k], obj_dict2[k]))
                    for k in diff.removed():
                        AssertionError('removed {}: {}'.format(k, obj_dict[k]))
                    for k in diff.added():
                        AssertionError('added {}: {}'.format(k, obj_dict2[k]))
                elif key in objclient.obj.scalarfields:
                    obj_dict2 = objclient.obj.from_apidict(server_changed_entities).get_apidict()
                    if server_changed_entities[key] != obj_dict2[key]:
                        AssertionError('changed {}: {}->{}'.format(key, server_changed_entities[key], obj_dict2[key]))


if __name__ == "__main__":
    unittest.main()



import unittest

from pynYNAB.ClientFactory import nYnabClientFactory
from tests.common_mock import MockConnection

connection = MockConnection()
factory = nYnabClientFactory('sqlite://')


class MockConnection(object):
    def __init__(self,user_id):
        self.user_id = user_id


class Args(object):
    budget_name = 'Test Budget'

    def __init__(self, user_id):
        self.connection = MockConnection(user_id)


class TestPersist(unittest.TestCase):
    def test_client_persist(self):
        cl1 = factory.create_client(Args('12345'), sync=False)
        cl2 = factory.create_client(Args('12345'), sync=False)
        self.assertEqual(cl1,cl2)
        cl3 = factory.create_client(Args('54231'), sync=False)
        self.assertNotEqual(cl1, cl3)


import logging

from pynYNAB.ClientFactory import nYnabClientFactory
from pynYNAB.schema import nYnabClient_

LOG = logging.getLogger(__name__)

class nYnabClient(object):
    def __new__(cls, *args, **kwargs):
        connection = kwargs.pop('nynabconnection', None)

        class Args(object):
            budget_name = kwargs.pop('budgetname', None)
            email = connection.email if hasattr(connection, 'email') else kwargs.pop('email', '')
            password = connection.password if hasattr(connection, 'password') else kwargs.pop('password', '')
            nynabconnection = connection

        passed = Args()
        factory = nYnabClientFactory()
        return factory.create_client(passed)



from __future__ import absolute_import

import datetime
import json
import unittest

from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Enum
from sqlalchemy import Integer
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from pynYNAB.ClientFactory import nYnabClientFactory
from pynYNAB.schema.Entity import Entity, ComplexEncoder, Base, AccountTypes
from pynYNAB.schema.budget import Account, Transaction, Subtransaction
from pynYNAB.schema.catalog import User
from pynYNAB.schema.roots import Budget
from pynYNAB.schema.types import AmountType
from tests.common_mock import MockConnection


class MyEntity(Base, Entity):
    greatfield = Column(Integer, default=2)


class CommonTest(unittest.TestCase):
    def setUp(self):
        engine = create_engine('sqlite://')

        Base.metadata.create_all(engine)
        self.Session = sessionmaker(bind=engine)
        self.session = self.Session()

factory = nYnabClientFactory()


class TestGetChangedEntities(CommonTest):
    def setUp(self):
        super(TestGetChangedEntities, self).setUp()
        self.obj = Budget()
        self.account = Account()
        self.obj.be_accounts = [self.account]
        self.obj.clear_changed_entities()
        self.account2 = Account(id=self.account)

        self.client = factory.create_client(budget_name='budgetname', nynabconnection = MockConnection(),sync=False)

    def testgetChangedEntities_add(self):
        added_account = Account()
        self.obj.be_accounts.append(added_account)
        changed_entities = self.obj.get_changed_entities()
        self.assertEqual(changed_entities, {'be_accounts': [added_account]})

    def testgetChangedEntities_addtransactionsubtransaction(self):
        added_transaction = Transaction()
        subtransaction1 = Subtransaction(entities_transaction=added_transaction)
        subtransaction2 = Subtransaction(entities_transaction=added_transaction)

        self.client.budget.be_transactions.append(added_transaction)
        self.client.budget.be_subtransactions.append(subtransaction1)
        self.client.budget.be_subtransactions.append(subtransaction2)

        self.client.session.commit()

        changed_entities = self.client.budget.get_changed_entities()
        self.assertIsInstance(changed_entities,dict)
        self.assertEqual(1, len(changed_entities.keys()))
        self.assertEqual('be_transaction_groups',list(changed_entities.keys())[0])
        transaction_groups = changed_entities['be_transaction_groups']

        self.assertEqual(1, len(transaction_groups))
        self.assertEqual(added_transaction, transaction_groups[0]['be_transaction'])

        self.assertIsNotNone(transaction_groups[0]['be_subtransactions'])
        try:
            self.assertItemsEqual([subtransaction1,subtransaction2], set(transaction_groups[0]['be_subtransactions']))
        except AttributeError:
            self.assertCountEqual([subtransaction1,subtransaction2], set(transaction_groups[0]['be_subtransactions']))


    def testgetChangedEntities_delete(self):
        self.obj.be_accounts.remove(self.account)
        changed_entities = self.obj.get_changed_entities()
        deleted = self.account.copy()
        deleted.is_tombstone = True
        self.assertEqual(changed_entities, {'be_accounts': [deleted]})

    def testgetChangedEntities_modify(self):
        self.account.account_name = 'BLA'
        changed_entities = self.obj.get_changed_entities()
        self.assertEqual(changed_entities, {'be_accounts': [self.account]})

    def test_arraytype(self):
        user = User()
        user.feature_flags = ['featureA','feature1']
        self.session.add(user)
        self.session.commit()

        fetched_user = self.session.query(User).first()
        self.assertEqual(user,fetched_user)


class TestUpdateChangedEntities(CommonTest):
    def setUp(self):
        self.account = Account()

        self.client = factory.create_client(budgetname='budgetname', nynabconnection = MockConnection(),sync=False)
        self.client.budget.be_accounts = [self.account]
        self.account2 = self.account.copy()
        self.client.session.commit()

    def testupdateChangedEntities_add(self):
        new_account = Account()
        changed_entities = dict(
            be_accounts=[new_account]
        )
        self.client.budgetClient.update_from_changed_entities(changed_entities)
        self.assertEqual(len(self.client.budget.be_accounts), 2)
        self.assertIn(new_account, self.client.budget.be_accounts)

    def testupdateChangedEntities_delete(self):
        self.account2.is_tombstone = True
        changed_entities = dict(
            be_accounts=[self.account2]
        )
        self.client.budgetClient.update_from_changed_entities(changed_entities)
        self.assertEqual(len(self.client.budget.be_accounts), 0)

    def testupdateChangedEntities_modify(self):
        self.account2 = self.account.copy()
        self.account2.note = 'note'
        changed_entities = dict(
            be_accounts=[self.account2]
        )

        self.client.budgetClient.update_from_changed_entities(changed_entities)
        self.assertEqual(len(self.client.budget.be_accounts), 1)
        acc = self.client.budget.be_accounts[0]
        self.assertEqual(acc, self.account2)


class OtherTests(CommonTest):
    def testEntityjson(self):
        obj = MyEntity()
        self.session.add(obj)
        self.session.commit()

        jsonroundtrip = json.loads(json.dumps(obj, cls=ComplexEncoder))
        self.assertEqual({'id': str(obj.id), 'greatfield': 2, 'is_tombstone': False}, jsonroundtrip)

    def testequality(self):
        tr1 = Transaction(id='t')
        tr2 = Transaction(id='t')
        self.assertTrue(tr1 == tr2)

        tr1 = Transaction()
        tr2 = Transaction()
        self.assertFalse(tr1 == tr2)

    def testrepr(self):
        tr1 = Transaction(id='t')
        self.assertEqual(tr1.__repr__(),tr1.__str__())

    def testappend(self):
        obj = Budget()
        account = Account()
        obj.be_accounts.append(account)
        self.assertEqual(len(obj.be_accounts), 1)
        self.assertEqual(list(obj.be_accounts)[-1], account)

    def testappendBad(self):
        obj = Budget()
        transaction = Transaction()

        def testappend():
            obj.be_accounts.append(transaction)

        self.assertRaises(ValueError, testappend)

    def test_str(self):
        # tests no exceptions when getting the string representation of some entities
        obj = Transaction()
        obj.__str__()
        obj.__unicode__()

        obj2 = Budget()
        obj2.be_accounts.__str__()

    def jsondefault(self):
        encoded = json.dumps('test', cls=ComplexEncoder)
        self.assertEqual(encoded, 'test')


class TestOthers(unittest.TestCase):
    def test_copy(self):
        obj = Account()
        objc = obj.copy()
        self.assertEqual(obj.id, objc.id)
        self.assertEqual(obj.get_dict(), objc.get_dict())


class DummyEntity(Base, Entity):
    account_type = Column(Enum(AccountTypes), default=AccountTypes.undef)
    date = Column(Date)
    balance = Column(AmountType)


class TestApiDict(unittest.TestCase):
    def test_input(self):
        inputdict=dict(
            id='dummyid/ca85f126-04a1-4196-bbeb-a77acec4b28e',
            account_type='Checking',
            date='2016-10-01',
            balance=1000,
            is_tombstone=False
        )
        dummy = DummyEntity.from_apidict(inputdict)
        self.assertEqual(dummy.id,inputdict['id'])
        self.assertEqual(dummy.account_type,AccountTypes.Checking)
        self.assertEqual(dummy.date, datetime.date(year=2016,month=10,day=1))
        self.assertEqual(dummy.balance,1)

    def test_output(self):
        inputentity = DummyEntity(
            account_type = AccountTypes.Checking,
            date = datetime.date(year=2016,month=10,day=1),
            balance = 1
        )
        entitydict = inputentity.get_apidict()
        self.assertEqual(entitydict['id'], str(inputentity.id))
        self.assertEqual(entitydict['account_type'], 'Checking')
        self.assertEqual(entitydict['date'], '2016-10-01')
        self.assertEqual(entitydict['balance'], 1000)



import os
import re
from lxml import html
import requests
import jsbeautifier
response = requests.get('http://app.youneedabudget.com')
with open('index.html', 'w', encoding='utf-8') as file_before:
    file_before.write(response.text)
parsed = html.fromstring(response.text)
for src in parsed.xpath('//script/@src'):
    url_src = str(src)
    file = url_src.rsplit('/',1)[-1]
    if file.startswith('before.'):
        before_response = requests.get(str(src))
        before_script = jsbeautifier.beautify(before_response.text)
        with open(os.path.join('web_app','before.js'),'w',encoding='utf-8') as file_before:
            file_before.write(before_script)
        regex1 = re.compile('\s*(\d)\:\s\"appmain\"')

        regex2=None
        for line in before_script.split('\n'):
            if regex1.match(line):
                idx = regex1.match(line).groups()[0]
                regex2 = re.compile('\s*%s\:\s\"(.*)\"' % idx)
            if regex2 is not None and regex2.match(line):
                test = regex2.match(line).groups()[0]
                if test!='appmain':
                    random_id = test
                    break
        url_appmain = '/'.join(url_src.rsplit('/',1)[:-1]+['appmain.'+random_id+'.js'])
        appmain_response = requests.get(url_appmain)
        appmain_script = jsbeautifier.beautify(appmain_response.text)
        with open(os.path.join('web_app','appmain.js'),'w',encoding='utf-8') as file_appmain:
            file_appmain.write(appmain_script)

    if file.startswith('index.'):
        script_response = requests.get(str(src))
        index_script = jsbeautifier.beautify(script_response.text)
        with open(os.path.join('web_app','index.js'),'w',encoding='utf-8') as file_before:
            file_before.write(index_script)


pass


import argparse
import inspect
import os
from getpass import getpass

import configargparse
import logging

import sys
from appdirs import AppDirs

from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.scripts.csvimport import do_csvimport
from pynYNAB.scripts.ofximport import do_ofximport

configfile = 'ynab.conf'
if not os.path.exists(configfile):
    myAppdir = AppDirs('pynYNAB').user_config_dir
    configfile = os.path.join(myAppdir, configfile)

LOG = logging.getLogger(__name__)

parser = configargparse.getArgumentParser('pynYNAB', default_config_files=[configfile],
                                          add_env_var_help=True,
                                          add_config_file_help=True,
                                          auto_env_var_prefix='NYNAB_')

parser.add_argument('--email', metavar='Email', type=str, required=False,
                    help='The Email User ID for nYNAB')
parser.add_argument('--password', metavar='Password', type=str, required=False,
                    help='**insecure** The Password for nYNAB, just omit it to enter it securely in a prompt')
parser.add_argument('--budgetname', metavar='BudgetName', type=str, required=False,
                    help='The nYNAB budget to use')


class classproperty(object):
    def __init__(self, f):
        self.f = f
    def __get__(self, obj, owner):
        return self.f(owner)


class MainCommands(object):
    def __init__(self):
        parser = argparse.ArgumentParser(
            description='nYnab-CLI using the python API',
            usage='''nynab <command> [<args>]

            ''')
        argcommands = parser.add_argument('command', help='Subcommand to run',choices=['csvimport','ofximport'])
        parser.usage += 'commands : '+','.join(argcommands.choices)
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print('Unrecognized command')
            parser.print_help()
            exit(1)
        if args.password is None:
            args.password = getpass('YNAB password:')
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    @classproperty
    def csvimport_parser(cls):
        csv_parser = argparse.ArgumentParser(parents=[parser],add_help=False)
        csv_parser.description = inspect.getdoc(cls.csvimport)
        csv_parser.add_argument('csvfile', metavar='CSVpath', type=str,
                                      help='The CSV file to import')
        csv_parser.add_argument('schema', metavar='schemaName', type=str,
                                      help='The CSV schema to use (see csv_schemas directory)')
        csv_parser.add_argument('accountname', metavar='AccountName', type=str, nargs='?',
                                      help='The nYNAB account name  to use')
        csv_parser.add_argument('-import-duplicates', action='store_true',
                                      help='Forces the import even if a duplicate (same date, account, amount, memo, payee) is found')
        return csv_parser


    @classmethod
    def csvimport(cls):
        """Manually import a CSV into a nYNAB budget"""
        print('pynYNAB CSV import')

        args = cls.csvimport_parser.parse_args()
        verify_common_args(args)

        if not os.path.exists(args.csvfile):
            LOG.error('input CSV file does not exist')
            exit(-1)

        client = clientfromargs(args)
        delta = do_csvimport(args,client)
        client.push(expected_delta=delta)

    @classproperty
    def ofximport_parser(cls):
        ofx_parser = argparse.ArgumentParser(parents=[parser],add_help=False)
        ofx_parser.description = inspect.getdoc(cls.ofximport)
        ofx_parser.add_argument('ofxfile', metavar='OFXPath', type=str,
                            help='The OFX file to import')
        return ofx_parser

    @classmethod
    def ofximport(cls):
        """Manually import an OFX into a nYNAB budget"""
        print('pynYNAB OFX import')

        args = cls.ofximport_parser.parse_args()
        verify_common_args(args)
        client = clientfromargs(args)
        delta = do_ofximport(args,client)
        client.push(expected_delta=delta)


def verify_common_args(args):
    if args.email is None:
        LOG.error('No email user ID provided, please specify it at the command line or in %s' % (configfile,))
        exit(-1)
    if args.password is None:
        LOG.error('No password provided, please specify it at the command line or in %s' % (configfile,))
        exit(-1)
    if args.budgetname is None:
        LOG.error('No budget name provided, please specify it at the command line or in %s' % (configfile,))
        exit(-1)


def main():
    MainCommands()


import logging
from abc import abstractproperty,ABCMeta

from pynYNAB.schema import fromapi_conversion_functions_table

LOG = logging.getLogger(__name__)



class RootObjClient():
    __metaclass__ = ABCMeta

    @abstractproperty
    def extra(self):
        return {}

    @abstractproperty
    def opname(self):
        return ''

    def __init__(self, obj, client):
        self.obj = obj
        self.client = client
        self.connection = client.connection
        self.session = client.session
        self.server_entities = {}
        self.synced = False

    def update_from_api_changed_entitydicts(self, changed_entitydicts, update_keys=None):
        if update_keys is None:
            update_keys = list(self.obj.listfields.keys())
        else:
            update_keys = [k for k in update_keys if k in self.obj.listfields]
        modified_entitydicts = {}
        for listfield_name in update_keys:
            newlist = []
            if changed_entitydicts[listfield_name] is not None:
                for entitydict in changed_entitydicts[listfield_name]:
                    newlist.append(self.obj.listfields[listfield_name].from_apidict(entitydict))
            modified_entitydicts[listfield_name] = newlist
        for scalarfield_name in self.obj.scalarfields:
            if scalarfield_name in changed_entitydicts:
                typ = self.obj.scalarfields[scalarfield_name]
                conversion_function = fromapi_conversion_functions_table.get(typ, lambda t, x: x)
                modified_entitydicts[scalarfield_name] = conversion_function(typ, changed_entitydicts[scalarfield_name])
        self.update_from_changed_entities(modified_entitydicts)

    def update_from_changed_entitydict(self, changed_entitiydicts):
        for name in changed_entitiydicts:
            value = changed_entitiydicts[name]
            if not isinstance(value, list):
                continue

            for incoming_obj_dict in value:
                incoming_obj_dict['parent_id'] = self.obj.id
                current_obj = self.session.query(self.obj.listfields[name]).get(incoming_obj_dict['id'])
                if current_obj is not None:
                    if incoming_obj_dict['is_tombstone']:
                        self.session.delete(current_obj)
                        continue
                else:
                    incoming_obj = self.obj.listfields[name].from_dict(incoming_obj_dict)
                    self.session.merge(incoming_obj)
        self.session.commit()
        self.session.expire(self.obj)
        pass

    def update_from_changed_entities(self, changed_entities):
        for name in changed_entities:
            value = changed_entities[name]
            if not isinstance(value, list):
                continue
            list_of_entities = getattr(self.obj, name)
            for incoming_obj in value:
                current_obj = self.session.query(self.obj.listfields[name]).get(incoming_obj.id)
                if current_obj is not None:
                    if incoming_obj.is_tombstone:
                        self.session.delete(current_obj)
                    else:
                        if current_obj not in list_of_entities:
                            current_obj.parent = self.obj
                        else:
                            for field in current_obj.scalarfields:
                                incoming = getattr(incoming_obj, field)
                                present = getattr(current_obj, field)
                                if present != incoming:
                                    setattr(current_obj, field, incoming)
                                    pass
                            pass
                else:
                    if not incoming_obj.is_tombstone:
                        self.session.add(incoming_obj)
                        incoming_obj.parent = self.obj
        self.session.commit()
        pass

    def update_from_sync_data(self, sync_data, update_keys=None):
        self.update_from_api_changed_entitydicts(sync_data['changed_entities'],update_keys)


    def sync(self, update_keys=None):
        if self.connection is None:
            return
        sync_data = self.get_sync_data_obj()

        self.server_entities[self.opname] = sync_data['changed_entities']
        LOG.debug('server_knowledge_of_device ' + str(sync_data['server_knowledge_of_device']))
        LOG.debug('current_server_knowledge ' + str(sync_data['current_server_knowledge']))
        self.update_from_sync_data(sync_data,update_keys)
        self.session.commit()
        self.obj.clear_changed_entities()

        server_knowledge_of_device = sync_data['server_knowledge_of_device']
        current_server_knowledge = sync_data['current_server_knowledge']

        change = current_server_knowledge - self.obj.knowledge.device_knowledge_of_server
        if change > 0:
            LOG.debug('Server knowledge has gone up by ' + str(
                change) + '. We should be getting back some entities from the server')
        if  self.obj.knowledge.current_device_knowledge < server_knowledge_of_device:
            if  self.obj.knowledge.current_device_knowledge != 0:
                LOG.error('The server knows more about this device than we know about ourselves')
            self.obj.knowledge.current_device_knowledge = server_knowledge_of_device
        self.obj.knowledge.device_knowledge_of_server = current_server_knowledge

        LOG.debug('current_device_knowledge %s' %  self.obj.knowledge.current_device_knowledge)
        LOG.debug('device_knowledge_of_server %s' % self.obj.knowledge.device_knowledge_of_server)
        self.synced = True

    def push(self, update_from_sync_data=True, update_keys=None):
        changed_entities = self.obj.get_changed_apidict()
        request_data = dict(starting_device_knowledge=self.client.starting_device_knowledge,
                            ending_device_knowledge=self.client.ending_device_knowledge,
                            device_knowledge_of_server=self.obj.knowledge.device_knowledge_of_server,
                            changed_entities=changed_entities)
        request_data.update(self.extra)

        def validate():
            self.session.commit()
            self.obj.clear_changed_entities()
        if self.connection is not None:
            sync_data = self.connection.dorequest(request_data, self.opname)
            LOG.debug('server_knowledge_of_device ' + str(sync_data['server_knowledge_of_device']))
            LOG.debug('current_server_knowledge ' + str(sync_data['current_server_knowledge']))
            if update_from_sync_data:
                self.update_from_sync_data(sync_data, update_keys)
                validate()

                server_knowledge_of_device = sync_data['server_knowledge_of_device']
                current_server_knowledge = sync_data['current_server_knowledge']

                change = current_server_knowledge - self.obj.knowledge.device_knowledge_of_server
                if change > 0:
                    LOG.debug('Server knowledge has gone up by ' + str(
                        change) + '. We should be getting back some entities from the server')
                if  self.obj.knowledge.current_device_knowledge < server_knowledge_of_device:
                    if  self.obj.knowledge.current_device_knowledge != 0:
                        LOG.error('The server knows more about this device than we know about ourselves')
                    self.obj.knowledge.current_device_knowledge = server_knowledge_of_device
                self.obj.knowledge.device_knowledge_of_server = current_server_knowledge

                LOG.debug('current_device_knowledge %s' %  self.obj.knowledge.current_device_knowledge)
                LOG.debug('device_knowledge_of_server %s' % self.obj.knowledge.device_knowledge_of_server)
        else:
            validate()

    def get_sync_data_obj(self):
        if self.connection is None:
            return

            # sync with disregard for knowledge, start from 0
        request_data = dict(starting_device_knowledge=self.client.starting_device_knowledge,
                            ending_device_knowledge=self.client.ending_device_knowledge,
                            device_knowledge_of_server=self.obj.knowledge.device_knowledge_of_server,
                            changed_entities={})

        request_data.update(self.extra)

        return self.connection.dorequest(request_data, self.opname)


import unittest

from pynYNAB.ClientFactory import nYnabClientFactory
from pynYNAB.schema.budget import SubCategory, Payee, MasterCategory
from pynYNAB.schema.catalog import BudgetVersion
from pynYNAB.schema.roots import Catalog, Budget


class MockConnection(object):
    def __init__(self):
        self.user_id = '1234'
        self.catalog = Catalog()
        self.budget = Budget()

    def dorequest(self, request_dic, opname):
        if opname == 'syncCatalogData':
            return {'changed_entities':{k:[] for k in self.catalog.listfields},'server_knowledge_of_device':0,'current_server_knowledge':123}
        if opname == 'syncBudgetData':
            return {'changed_entities':{k:[] for k in self.budget.listfields},'server_knowledge_of_device':0,'current_server_knowledge':123}

    def init_session(self):
        pass

    user_id = '1'
factory = nYnabClientFactory()

class TestCommonMock(unittest.TestCase):
    def setUp(self):
        self.client = factory.create_client(budgetname='TestBudget',nynabconnection=MockConnection(),sync=False)

        session = self.client.session

        budget_version = BudgetVersion(version_name='TestBudget')
        master_category = MasterCategory(name='master')
        subcategory = SubCategory(name='Immediate Income',
                                  internal_name='Category/__ImmediateIncome__',
                                  entities_master_category=master_category)
        payee = Payee(name='Starting Balance Payee',internal_name='StartingBalancePayee')
        session.add(master_category)
        session.add(subcategory)
        session.add(payee)

        self.client.catalog.ce_budget_versions.append(budget_version)
        self.client.budget.be_master_categories.append(master_category)
        self.client.budget.be_subcategories.append(subcategory)
        self.client.budget.be_payees.append(payee)
        session.commit()
        self.client.budget.clear_changed_entities()
        self.client.catalog.clear_changed_entities()

        self.client.budgetClient.device_knowledge_of_server = 0
        self.client.catalogClient.device_knowledge_of_server = 0

        self.client.budgetClient.current_device_knowledge = 0
        self.client.catalogClient.current_device_knowledge = 0

        pass



from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.__main__ import parser
from dotenv import load_dotenv,find_dotenv

from pynYNAB.connection import nYnabConnection
from pynYNAB.schema import Transaction
from pynYNAB.ClientFactory import nYnabClientFactory

load_dotenv(find_dotenv())

print('test_sync')
args = parser.parse_known_args()[0]

client = nYnabClientFactory().create_client(args)
client.sync()

for i in range(0,1000):
    client.budget.be_transactions.append(Transaction())
client.push(1000)
print('OK')



import json

from pynYNAB.Client import nYnabClient
from pynYNAB.ClientFactory import nYnabClientFactory
from pynYNAB.exceptions import NoBudgetNameException
from pynYNAB.schema.catalog import BudgetVersion
from tests.common_mock import TestCommonMock


currency_format = dict(
    iso_code='USD',
    example_format='123,456.78',
    decimal_digits=2,
    decimal_separator='.',
    symbol_first=True,
    group_separator=',',
    currency_symbol='$',
    display_symbol=True
)
date_format = dict(
    format='MM/DD/YYYY'
)


class MockConnection2(object):
    user_id='12345'

factory = nYnabClientFactory('sqlite://')


class TestOperations(TestCommonMock):
    def test_create_budget(self):
        class MockConnection2(object):
            def dorequest(this,request_dic, opname):
                self.assertEqual(opname, opname)
                self.assertEqual(request_dic['currency_format'], json.dumps(currency_format))
                self.assertEqual(request_dic['date_format'], json.dumps(date_format))
            user_id='1234'

        self.client = factory.create_client(budgetname='', nynabconnection=MockConnection2(), sync=False)
        self.client.create_budget(budget_name='New Budget')

    def test_client_nobudget(self):
        def create_client_no_budget():
            client = nYnabClient()
        self.assertRaises(NoBudgetNameException, create_client_no_budget)

    def test_select_budget(self):
        client = factory.create_client(budget_name='',nynabconnection=MockConnection2(), sync=False)
        budget_version1 = BudgetVersion(version_name='TestBudget')
        budget_version2 = BudgetVersion(version_name='NewTestBudget')
        client.catalog.ce_budget_versions= [budget_version1, budget_version2]

        client.select_budget(budget_name='NewTestBudget')
        self.assertEqual(budget_version2.id, client.budget_version_id)

    def test_create_client(self):
        class Args(object):
            nynabconnection = MockConnection2()
            budgetname = 'budgetname'
        client = factory.create_client(Args, sync=False)
        self.assertEqual(Args.nynabconnection,client.connection)
        self.assertEqual(Args.nynabconnection, client.catalogClient.connection)
        self.assertEqual(Args.nynabconnection, client.budgetClient.connection)
        self.assertEqual(Args.budgetname, client.budget_name)

    def test_create_client_nynabconnectionparameter(self):
        class Args(object):
            nynabconnection = MockConnection2()
            budgetname = 'budgetname'

        client = factory.create_client(Args, sync=False)
        self.assertEqual(Args.nynabconnection, client.connection)
        self.assertEqual(Args.budgetname, client.budget_name)
        self.assertEqual('sqlite://', str(client.session.bind.url))



from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship

from pynYNAB.schema.Entity import Entity, Base, RootEntity
from pynYNAB.schema.types import ArrayType


class CatalogEntity(Entity):
    @declared_attr
    def parent_id(self):
        return Column(ForeignKey('catalog.id'))

    @declared_attr
    def parent(self):
        return relationship('Catalog')


class CatalogBudget(Base, CatalogEntity):
    budget_name = Column(String)
    created_at = Column(DateTime)


class User(Base, CatalogEntity):
    username = Column(String)
    trial_expires_on = Column(String)
    email = Column(String)
    feature_flags = Column(ArrayType)
    is_subscribed = Column(Boolean)


class UserSetting(Base, CatalogEntity):
    setting_name = Column(String)
    user_id = Column(ForeignKey('user.id'))
    user = relationship('User', foreign_keys=user_id, backref='settings')
    setting_value = Column(String)


class UserBudget(Base, CatalogEntity):
    budget_id = Column(ForeignKey('catalogbudget.id'))
    budget = relationship('CatalogBudget')

    user_id = Column(ForeignKey('user.id'))
    user = relationship('User', foreign_keys=user_id, backref='budgets')
    permissions = Column(String)


class BudgetVersion(Base, CatalogEntity):
    date_format = Column(String)
    last_accessed_on = Column(String)
    currency_format = Column(String)
    budget_id = Column(ForeignKey('catalogbudget.id'))
    budget = relationship('CatalogBudget', foreign_keys=budget_id)
    version_name = Column(String)
    source = Column(String)



import unittest

from pynYNAB.Client import WrongPushException
from pynYNAB.schema.Client import nYnabClient_
from pynYNAB.schema.budget import Transaction


class TestWrongPush(unittest.TestCase):
    def TestWrongPush(self):
        client = nYnabClient_(budgetname='Test')
        client.budget.be_transactions.append(Transaction())
        client.budget.be_transactions.append(Transaction())
        self.assertRaises(WrongPushException,lambda: client.push(expected_delta=1))



import os
from datetime import datetime
from tempfile import gettempdir

import configargparse

from pynYNAB.schema.budget import Transaction
from pynYNAB.scripts.csvimport import do_csvimport
from pynYNAB.utils import get_or_create_account, get_or_create_payee
from test_live.common import needs_account
from tests.common_mock import TestCommonMock


class TestCsv(TestCommonMock):
    def getTr(self, date, payee, amount, memo, account):
        imported_date = datetime.now().date()
        return Transaction(
            entities_account_id=get_or_create_account(self.client, account).id,
            date=date,
            entities_payee_id=get_or_create_payee(self.client, payee).id,
            imported_payee=payee,
            source='Imported',
            memo=memo,
            amount=amount,
            cash_amount=amount,
            imported_date=imported_date
        )

    @needs_account('Credit')
    def test_duplicate(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        args.schema = 'example'
        args.csvfile = os.path.join(gettempdir(), 'data.csv')
        args.accountname = None
        args.import_duplicates = False
        args.logginglevel = 'debug'

        content = """Date,Payee,Amount,Memo,Account
2016-02-01,Super Pants Inc.,-20,Buying pants,Credit
"""
        with open(args.csvfile, mode='w') as f:
            f.writelines(content)

        transaction = self.getTr(datetime(year=2016, month=2, day=1).date(), 'Super Pants Inc.', -20, 'Buying pants',
                                 'Credit')
        self.client.budget.be_transactions.append(transaction)
        delta = do_csvimport(args, self.client)
        self.assertEqual(delta, 0)

    @needs_account('Cash')
    def test_duplicateForced(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        args.schema = 'example'
        args.csvfile = os.path.join(gettempdir(), 'data.csv')
        args.accountname = None
        args.import_duplicates = True
        args.logginglevel = 'debug'

        content = """Date,Payee,Amount,Memo,Account
2016-02-01,Super Pants Inc.,-20,Buying pants,Cash
"""
        with open(args.csvfile, mode='w') as f:
            f.writelines(content)

        transaction = self.getTr(datetime(year=2016, month=2, day=1).date(), 'Super Pants Inc.', -20, 'Buying pants',
                                 'Cash')

        self.client.budget.be_transactions.append(transaction)
        delta = do_csvimport(args, self.client)
        self.assertEqual(delta, 1)

        self.assertEqual(sum(1 for tr in self.client.budget.be_transactions if tr.key2 == transaction.key2), 2)

    @needs_account('Cash')
    @needs_account('Checking Account')
    @needs_account('Savings')
    def test_import(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        args.schema = 'example'
        args.csvfile = os.path.join(gettempdir(), 'data.csv')
        args.accountname = None
        args.import_duplicates = False
        args.logginglevel = 'debug'

        content = """Date,Payee,Amount,Memo,Account
2016-02-01,Super Pants Inc.,-20,Buying pants,Cash
2016-02-02,Thai Restaurant,-10,Food,Checking Account
2016-02-03,,10,Saving!,Savings
        """
        with open(args.csvfile, mode='w') as f:
            f.writelines(content)

        transactions = [
            self.getTr(datetime(year=2016, month=2, day=1).date(), 'Super Pants Inc.', -20, 'Buying pants', 'Cash'),
            self.getTr(datetime(year=2016, month=2, day=2).date(), 'Thai Restaurant', -10, 'Food', 'Checking Account'),
            self.getTr(datetime(year=2016, month=2, day=3).date(), '', 10, 'Saving!', 'Savings'),
        ]

        do_csvimport(args, self.client)
        for transaction in transactions:
            self.assertIn(transaction.key2, [tr.key2 for tr in self.client.budget.be_transactions])



class BudgetNotFound(Exception):
    pass


class WrongPushException(Exception):
    def __init__(self, expected_delta, delta):
        self.expected_delta = expected_delta
        self.delta = delta

    string = 'tried to push a changed_entities with %d entities while we expected %d entities'

    @property
    def msg(self):
        return self.string % (self.delta, self.expected_delta)

    def __repr__(self):
        return self.msg


class NoBudgetNameException(ValueError):
    def __init__(self):
        super(NoBudgetNameException,self).__init__('you should pass a budget_name ')


# coding=utf-8
import unittest
from dotenv import load_dotenv,find_dotenv

from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.__main__ import parser
from pynYNAB.schema import DictDiffer
from pynYNAB.schema.Entity import fromapi_conversion_functions_table
from pynYNAB.schema.types import AmountType
## these test cases are live, but designed to be very short in time !

load_dotenv(find_dotenv())


class LiveTests2(unittest.TestCase):
    def test_roundtrip(self):
        args = parser.parse_known_args()[0]

        # 1. gets sync data from server
        # 2. tests that to_api(from_api(data)) is the same thing

        client = clientfromargs(args, sync=False)
        sync_data = client.catalogClient.get_sync_data_obj()
        budget_version_id = next(d['id'] for d in sync_data['changed_entities']['ce_budget_versions'] if
                                 d['version_name'] == args.budgetname)
        client.budget_version_id = budget_version_id

        for objclient in (client.budgetClient, client.catalogClient):
            sync_data = objclient.get_sync_data_obj()
            server_changed_entities = sync_data['changed_entities']

            for key in server_changed_entities:
                if key in objclient.obj.listfields:
                    if len(server_changed_entities[key]) == 0:
                        continue
                    obj_dict = server_changed_entities[key][0]
                    typ = objclient.obj.listfields[key]
                    obj_dict2 = typ.from_apidict(obj_dict).get_apidict()

                    diff = DictDiffer(obj_dict2, obj_dict)
                    for k in diff.changed():
                        AssertionError('changed {}: {}->{}'.format(k, obj_dict[k], obj_dict2[k]))
                    for k in diff.removed():
                        AssertionError('removed {}: {}'.format(k, obj_dict[k]))
                    for k in diff.added():
                        AssertionError('added {}: {}'.format(k, obj_dict2[k]))
                elif key in objclient.obj.scalarfields:
                    obj_dict2 = objclient.obj.from_apidict(server_changed_entities).get_apidict()
                    if server_changed_entities[key] != obj_dict2[key]:
                        AssertionError('changed {}: {}->{}'.format(key, server_changed_entities[key], obj_dict2[key]))

test_budget_name = 'Test Budget - Dont Remove'

# this test cases expect that
# a budget named "Test Budget" exists
# in it, there is an account named "Account"
# in it there is a transaction date 27/01/2017 that has inflow == 12.34 € with a memo "TEST TRANSACTION"
# We check the API dict we fetch when syncing budget
class LiveTestBudget(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(LiveTestBudget, self).__init__(*args, **kwargs)
        self.transaction = None
        self.client = None

    def setUp(self):
        args = parser.parse_known_args()[0]
        args.budgetname = test_budget_name
        self.client = clientfromargs(args, sync=False)
        self.client.catalogClient.sync()
        self.client.select_budget(test_budget_name)

    def test_api_scaling_is_ok(self):
        sync_data = self.client.budgetClient.get_sync_data_obj()
        server_entities = sync_data['changed_entities']
        transactions = server_entities['be_transactions']
        amount = None
        for transaction in transactions:
            if transaction['memo'] == 'TEST TRANSACTION':
                amount = fromapi_conversion_functions_table[AmountType](AmountType,transaction['amount'])
        self.assertEqual(12.34,amount)


import json
import unittest

from sqlalchemy import create_engine

from pynYNAB.ClientFactory import nYnabClientFactory


class TestSyncMock(unittest.TestCase):
    def test_sync_mock(self):
        class MockConnection(object):
            user_id='1'

            def dorequest(self, request_data, opname):
                with open(opname+'.json','r') as f:
                    data = json.load(f)
                    return data
        engine= create_engine('sqlite://',echo=True)

        client = nYnabClientFactory(engine=engine).create_client(budgetname='Test Budget',
                                                                 connection = MockConnection(),
                                                                 sync=True)


    def test_update_obj(self):
        pass




from datetime import datetime

from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.schema.budget import Transaction
from pynYNAB.__main__ import parser
from dotenv import load_dotenv,find_dotenv

from pynYNAB.utils import get_or_create_payee

load_dotenv(find_dotenv())

print('test_sync')
args = parser.parse_known_args()[0]
client = clientfromargs(args)
client.sync()

accounts = {v.account_name:v for v in client.budget.be_accounts}
payees = {p.name:p for p in client.budget.be_payees}
listaccounts = list(accounts.values())
account0 = listaccounts[0]
account1 = listaccounts[1]

transactions = []

account_id = account0.id
transfer_id = account1.id

# Payees:  "Transfer : account_name"
account_payee_id = payees['Transfer : ' + account1.account_name].id
transfer_payee_id = payees['Transfer : ' + account0.account_name].id

imported_date = datetime.now().date()

payee_name = 'Payee'
payee = get_or_create_payee(client, payee_name)
payee_id = payee.id

amount = 10
date = datetime(2016,10,5).date()


transaction_1 = Transaction(
    entities_account_id=account_id,
    entities_payee_id=account_payee_id,
    amount=amount,
    date=date,
    imported_date=imported_date,
    imported_payee=payee_name,
    source='Matched',
    cash_amount=amount
)

transaction_orig = Transaction(
    entities_account_id=account_id,
    amount=amount,
    date=date,
    entities_payee_id=payee_id,
    imported_date=imported_date,
    memo="",
    source='matched_import'
)

# match
transaction_1.matched_transaction_id = transaction_orig.id
transaction_orig.matched_transaction_id = transaction_1.id

transactions.append(transaction_orig)
transactions.append(transaction_1)

client.add_transactions(transactions)


from sqlalchemy import Boolean, Column, String, Integer
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship

from pynYNAB import KeyGenerator
from pynYNAB.schema import Base

from pynYNAB.schema.Entity import Base, RootEntity
from pynYNAB.schema.budget import TransactionGroup


class Catalog(Base, RootEntity):
    ce_user_budgets = relationship('UserBudget')
    ce_user_settings = relationship('UserSetting')
    ce_budget_versions = relationship('BudgetVersion')
    ce_users = relationship('User')
    ce_budgets = relationship('CatalogBudget')


class Budget(Base, RootEntity):
    be_transactions = relationship('Transaction')
    be_master_categories = relationship('MasterCategory')
    be_settings = relationship('Setting')
    be_monthly_budget_calculations = relationship('MonthlyBudgetCalculation')
    be_account_mappings = relationship('AccountMapping')
    be_subtransactions = relationship('Subtransaction')
    be_scheduled_subtransactions = relationship('ScheduledSubtransaction')
    be_monthly_budgets = relationship('MonthlyBudget')
    be_subcategories = relationship('SubCategory')
    be_payee_locations = relationship('PayeeLocation')
    be_account_calculations = relationship('AccountCalculation')
    be_monthly_account_calculations = relationship('MonthlyAccountCalculation')
    be_monthly_subcategory_budget_calculations = relationship('MonthlySubcategoryBudgetCalculation')
    be_scheduled_transactions = relationship('ScheduledTransaction')
    be_payees = relationship('Payee')
    be_monthly_subcategory_budgets = relationship('MonthlySubcategoryBudget')
    be_payee_rename_conditions = relationship('PayeeRenameCondition')
    be_accounts = relationship('Account')
    last_month = Column(Date)
    first_month = Column(Date)

    knowledge = relationship('Knowledge')

    def get_changed_entities(self):
        changed_entities = super(Budget, self).get_changed_entities()
        if 'be_transactions' in changed_entities:
            changed_entities['be_transaction_groups'] = []
            for tr in changed_entities.pop('be_transactions'):
                subtransactions = []
                if 'be_subtransactions' in changed_entities:
                    for subtransaction in changed_entities['be_subtransactions']:
                        if subtransaction.entities_transaction_id == tr.id:
                            subtransactions.append(subtransaction)
                    for subtransaction in subtransactions:
                        changed_entities['be_subtransactions'].remove(subtransaction)
                if not subtransactions:
                    subtransactions = None
                group = TransactionGroup(
                    id=tr.id,
                    be_transaction=tr,
                    be_subtransactions=subtransactions,
                    be_matched_transaction=None)
                changed_entities['be_transaction_groups'].append(group)
        if changed_entities.get('be_subtransactions') is not None:
            del changed_entities['be_subtransactions']
        return changed_entities


class Knowledge(Base):
    __tablename__ = 'knowledge'

    obj_id = Column(String,primary_key=True,default=KeyGenerator.generateuuid)
    current_device_knowledge = Column(Integer,default=0)
    device_knowledge_of_server = Column(Integer,default=0)


import random
import unittest
from datetime import datetime
from functools import wraps

from pynYNAB import KeyGenerator
from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.schema.Entity import AccountTypes
from pynYNAB.schema.budget import Account, Payee
from pynYNAB.__main__ import parser


def util_add_account(client, account_name=None):
    if account_name is None:
        account_name = str(KeyGenerator.generateuuid())
    account = Account(
        account_type=random.choice(list(AccountTypes)),
        account_name=account_name
    )

    client.add_account(account, balance=random.randint(-10, 10), balance_date=datetime.now())
    return account


def needs_account(account_name=None):
    def decorator(fn):
        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            for account in self.client.budget.be_accounts:
                if account_name is None or account.account_name == account_name:
                    self.account = account
                    fn(self, *args, **kwargs)
                    return
            self.account = util_add_account(self.client,account_name)
            fn(self, *args, **kwargs)
            return

        return wrapped

    return decorator


class CommonLive(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(CommonLive, self).__init__(*args, **kwargs)
        self.account = None
        self.budget = None
        self.transaction = None
        self.client = None

    def reload(self):
        # parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        self.client = clientfromargs(args)

    def setUp(self):
        self.reload()





import json
import os
import tempfile
import unittest
from datetime import datetime

import configargparse

from pynYNAB.schema.budget import Transaction
from pynYNAB.scripts.ofximport import do_ofximport
from pynYNAB.utils import get_or_create_account, get_or_create_payee
from tests.common_mock import TestCommonMock


class TestOFX(TestCommonMock):
    def testiimport(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        args.ofxfile = os.path.join(tempfile.gettempdir(), 'data.ofx')
        args.logginglevel = 'debug'

        content = """OFXHEADER:100
DATA:OFXSGML
VERSION:102
SECURITY:NONE
ENCODING:USASCII
CHARSET:1252
COMPRESSION:NONE
OLDFILEUID:NONE
NEWFILEUID:NONE
<OFX>
    <SIGNONMSGSRSV1>
    <SONRS>
        <STATUS>
            <CODE>0
            <SEVERITY>INFO
        </STATUS>
        <DTSERVER>20130313133728
        <LANGUAGE>FRA
    </SONRS>
    </SIGNONMSGSRSV1>
    <BANKMSGSRSV1>
        <STMTTRNRS>
            <TRNUID>29939615002
            <STATUS>
                <CODE>0
                <SEVERITY>INFO
            </STATUS>
            <STMTRS>
                <CURDEF>EUR
                    <BANKACCTFROM>
                    <BANKID>11706
                    <BRANCHID>41029
                    <ACCTID>29939615002
                    <ACCTTYPE>CHECKING
                </BANKACCTFROM>
                <BANKTRANLIST>
                    <DTSTART>20130128000000
                    <DTEND>20130314235959
                    <STMTTRN>
                        <TRNTYPE>CHECK
                        <DTPOSTED>20130312
                        <TRNAMT>-491.09
                        <FITID>13071099780237330004
                        <CHECKNUM>0003445
                        <NAME>CHEQUE
                        <MEMO>CHEQUE
                    </STMTTRN>
                </BANKTRANLIST>
                <LEDGERBAL>
                    <BALAMT>-6348.01
                    <DTASOF>20130312
                </LEDGERBAL>
                <AVAILBAL>
                <BALAMT>-6348.01
                <DTASOF>20130312
                </AVAILBAL>
            </STMTRS>
        </STMTTRNRS>
    </BANKMSGSRSV1>
</OFX>"""
        with open(args.ofxfile, 'w') as f:
            f.writelines(content)
        imported_date = datetime.now().date()

        testaccount = 'TEST'

        account = get_or_create_account(self.client,testaccount)

        key = '11706 41029 29939615002'
        account.note = 'great note key[%s]key' % key

        payee = get_or_create_payee(self.client,'CHEQUE')
        amount = -491.09

        transaction = Transaction(
            entities_account=account,
            date=datetime(year=2013, month=3, day=12).date(),
            entities_payee=payee,
            imported_payee=payee,
            source='Imported',
            check_number='0003445',
            memo='CHEQUE    13071099780237330004',
            amount=amount,
            credit_amount=0,
            imported_date=imported_date
        )

        delta = do_ofximport(args,self.client)
        self.assertEqual(delta,1,msg="should add exactly one transaction")
        self.assertIn(transaction.key2, [tr.key2 for tr in self.client.budget.be_transactions],
                      msg='couldnt find an imported transaction after ofx import')



# coding=utf-8
import json
import logging
from time import sleep

import requests
from requests.cookies import RequestsCookieJar

from pynYNAB.KeyGenerator import generateuuid
from pynYNAB.schema.Entity import ComplexEncoder
from pynYNAB.utils import rate_limited

LOG = logging.getLogger(__name__)


class NYnabConnectionError(Exception):
    pass


# noinspection PyPep8Naming
class nYnabConnection(object):
    urlCatalog = 'https://app.youneedabudget.com/api/v1/catalog'

    def init_session(self):
        self.session.cookies = RequestsCookieJar()

        self.session.headers['X-YNAB-Device-Id'] = self.id
        self.session.headers['User-Agent'] = 'python nYNAB API bot - rienafairefr rienafairefr@gmail.com'

        firstlogin = self.dorequest({"email": self.email, "password": self.password, "remember_me": True,
                                     "device_info": {"id": self.id}}, 'loginUser')
        if firstlogin is None:
            raise NYnabConnectionError('Couldnt connect with the provided email and password')
        self.sessionToken = firstlogin["session_token"]
        self.session.headers['X-Session-Token'] = self.sessionToken
        self.user_id = firstlogin['user']['id']

    def __init__(self, email, password):
        self.email = email
        self.password = password
        self.session = requests.Session()
        self.sessionToken = None
        self.id = str(generateuuid())
        self.lastrequest_elapsed = None

    @rate_limited(maxpersecond=5)
    def dorequest(self, request_dic, opname):
        """
        :param request_dic: a dictionary containing parameters for the request
        :param opname: An API operation name, available are: loginUser,getInitialUserData,logoutUser,createNewBudget,
        freshStartABudget,cloneBudget,deleteTombstonedBudgets,syncCatalogData,syncBudgetData,getInstitutionList,
        getInstitutionLoginFields,getInstitutionAccountList,registerAccountForDirectConnect,
        updateDirectConnectCredentials,poll,createFeedback,runSqlStatement
        :return: the dictionary of the result of the request
        """
        # Available operations :

        def errorout(message):
            LOG.error(message.replace(self.password,'********'))
            raise NYnabConnectionError(message)

        json_request_dict = json.dumps(request_dic, cls=ComplexEncoder)
        params = {u'operation_name': opname, 'request_data': json_request_dict}
        LOG.debug(('%s  ... %s ' % (opname, params)).replace(self.password,'********'))
        r = self.session.post(self.urlCatalog, params)
        self.lastrequest_elapsed = r.elapsed
        js = r.json()
        if r.status_code == 500:
            raise NYnabConnectionError('Uunrecoverable server error, sorry YNAB')
        if r.status_code != 200:
            LOG.debug('non-200 HTTP code: %s ' % r.text)
        if not 'error' in js:
            errorout('The server returned a json value without an error field')
        if js['error'] is None:
            return js
        error = js['error']
        if 'id' not in error:
            errorout('Error field %s without id returned from the API, %s' % (error, params))
        if error['id'] == 'user_not_found':
            errorout('API error, User Not Found')
        elif error['id'] == 'user_password_invalid':
            errorout('API error, User-Password combination invalid')
        elif error['id'] == 'request_throttled':
            LOG.debug('API Rrequest throttled')
            retryrafter = r.headers['Retry-After']
            LOG.debug('Waiting for %s s' % retryrafter)
            sleep(float(retryrafter))
            return self.dorequest(request_dic, opname)
        else:
            errorout('Unknown API Error \"%s\" was returned from the API when sending request (%s)' % (error['id'], params))




import logging
from functools import wraps

from sqlalchemy import Column, String, ForeignKey, Integer
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker

from pynYNAB.exceptions import BudgetNotFound, WrongPushException
from pynYNAB.schema import Base, Catalog, Budget, Knowledge, Payee, Transaction

LOG = logging.getLogger(__name__)


def operation(expected_delta):
    def operation_decorator(fn):
        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            fn(self, *args, **kwargs)
            LOG.debug('push after '+fn.__name__)
            self.push(expected_delta)
        return wrapped
    return operation_decorator


class nYnabClient_(Base):
    __tablename__ = "nynabclients"
    id = Column(String, primary_key=True)
    catalog_id = Column(ForeignKey('catalog.id'))
    catalog = relationship('Catalog')
    budget_id = Column(ForeignKey('budget.id'))
    budget = relationship('Budget')
    budget_version_id = Column(String)
    budget_name = Column(String)
    starting_device_knowledge = Column(Integer, default=0)
    ending_device_knowledge = Column(Integer, default=0)
    connection = None
    session = None

    @property
    def user_id(self):
        return self.id

    @property
    def online(self):
        return self.connection is not None

    def add_missing(self):
        self.catalog = Catalog()
        self.catalog.knowledge = Knowledge()
        self.budget = Budget()
        self.budget.knowledge = Knowledge()
        self.session.add(self.catalog)
        self.session.add(self.budget)
        self.session.commit()

    def init_internal_db(self):
        Base.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()

    def sync(self, update_keys=None):
        LOG.debug('Client.sync')

        self.catalogClient.sync(update_keys)
        self.select_budget(self.budget_name)
        self.budgetClient.sync(update_keys)

        if self.budget_version_id is None and self.budget_name is not None:
            raise BudgetNotFound()

    def push(self, expected_delta=1):
        # ending-starting represents the number of modifications that have been done to the data ?
        LOG.debug('Client.push')

        catalog_changed_entities = self.catalog.get_changed_apidict()
        budget_changed_entities = self.budget.get_changed_apidict()

        delta = sum(len(l) for k, l in catalog_changed_entities.items()) + \
            sum(len(l) for k, l in budget_changed_entities.items())

        if delta != expected_delta:
            raise WrongPushException(expected_delta, delta)

        if any(catalog_changed_entities) or any(budget_changed_entities):
            self.ending_device_knowledge = self.starting_device_knowledge + 1

        self.catalogClient.push()
        self.budgetClient.push()

        self.starting_device_knowledge = self.ending_device_knowledge
        self.session.commit()

    @operation(3)
    def add_account(self, account, balance, balance_date):
        payee = Payee(
            entities_account_id=account.id,
            enabled=True,
            auto_fill_subcategory_enabled=True,
            auto_fill_memo_enabled=False,
            auto_fill_amount_enabled=False,
            rename_on_import_enabled=False,
            name="Transfer : %s" % account.account_name
        )
        immediateincomeid = next(
            s.id for s in self.budget.be_subcategories if s.internal_name == 'Category/__ImmediateIncome__')
        startingbalanceid = next(p.id for p in self.budget.be_payees if p.internal_name == 'StartingBalancePayee')

        transaction = Transaction(
            accepted=True,
            amount=balance,
            entities_subcategory_id=immediateincomeid,
            cash_amount=0,
            cleared='Cleared',
            date=balance_date,
            entities_account_id=account.id,
            credit_amount=0,
            entities_payee_id=startingbalanceid,
            is_tombstone=False
        )

        self.budget.be_accounts.append(account)
        self.budget.be_payees.append(payee)
        self.budget.be_transactions.append(transaction)
        pass

    @operation(1)
    def delete_account(self, account):
        self.budget.be_accounts.remove(account)

    @operation(1)
    def add_transaction(self, transaction):
        self.budget.be_transactions.append(transaction)

    def add_transactions(self, transaction_list):
        @operation(len(transaction_list))
        def _add_transactions_method(self, tr_list):
            for tr in tr_list:
                self.budget.be_transactions.append(tr)

        return _add_transactions_method(transaction_list)

    @operation(1)
    def delete_transaction(self, transaction):
        self.budget.be_transactions.remove(transaction)

    @operation(1)
    def delete_budget(self, budget_name):
        for budget in self.catalog.ce_budgets:
            if budget.budget_name == budget_name:
                self.catalog.ce_budgets.remove(budget)

    def select_budget(self, budget_name):
        self.budget_version_id = None
        for budget_version in self.catalog.ce_budget_versions:
            if budget_version.version_name == budget_name:
                self.budget_version_id = budget_version.id
        if self.budget_version_id is None:
            raise BudgetNotFound()

    def create_budget(self, budget_name):
        import json
        currency_format = dict(
            iso_code='USD',
            example_format='123,456.78',
            decimal_digits=2,
            decimal_separator='.',
            symbol_first=True,
            group_separator=',',
            currency_symbol='$',
            display_symbol=True
        )
        date_format = dict(
            format='MM/DD/YYYY'
        )
        self.connection.dorequest(opname='CreateNewBudget',
                                  request_dic={
                                      "budget_name": budget_name,
                                      "currency_format": json.dumps(currency_format),
                                      "date_format": json.dumps(date_format)
                                  })




import datetime
import unittest

from sqlalchemy import Date

from pynYNAB.schema.Entity import fromapi_conversion_functions_table,toapi_conversion_functions_table
from pynYNAB.schema.types import AmountType


class TestConversionFunctions(unittest.TestCase):
    def doassert(self, t, v):
        toapi = toapi_conversion_functions_table[t]
        fromapi = fromapi_conversion_functions_table[t]
        converted_value = toapi(t,v)
        converted_back = fromapi(t,converted_value)
        self.assertEqual(v, converted_back)

    def test_scaling(self):
        value = 12.34
        self.doassert(AmountType,value)

    def test_date(self):
        value = datetime.date(2015,1,25)
        self.doassert(Date, value)




import logging
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from pynYNAB.ObjClient import RootObjClient
from pynYNAB.connection import nYnabConnection
from pynYNAB.exceptions import NoBudgetNameException, BudgetNotFound
from pynYNAB.schema import Base


LOG = logging.getLogger(__name__)


class CatalogClient(RootObjClient):
    @property
    def extra(self):
        return dict(user_id=self.client.user_id)

    opname = 'syncCatalogData'

    def __init__(self, client):
        super(CatalogClient, self).__init__(client.catalog, client)


class BudgetClient(RootObjClient):
    @property
    def extra(self):
        return dict(calculated_entities_included=False, budget_version_id=self.client.budget_version_id)

    opname = 'syncBudgetData'

    def __init__(self, client):
        super(BudgetClient, self).__init__(client.budget, client)


class nYnabClientFactory(object):
    def __init__(self, engine_url='sqlite://', engine=None):
        self.engine_url = engine_url
        if engine is None:
            self.engine = create_engine(engine_url)
        else:
            self.engine = engine

        Base.metadata.create_all(self.engine)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

    def create_client(self, args=None, sync=True, **kwargs):
        from pynYNAB.schema.Client import nYnabClient_
        from pynYNAB.Client import nYnabClient
        if args is None:
            class Arg(object):pass
            args = Arg()
        for k,v in kwargs.items():
            setattr(args,k,v)
        if hasattr(args,'budgetname'):
            setattr(args, 'budget_name', args.budgetname)

        if hasattr(args, 'nynabconnection'):
            setattr(args, 'connection', args.nynabconnection)

        if not hasattr(args, 'budget_name') or args.budget_name is None:
            raise NoBudgetNameException

        try:
            if not hasattr(args, 'connection'):
                connection = nYnabConnection(args.email, args.password)
                connection.init_session()
            else:
                connection = args.connection

            client_id = connection.user_id

            def postprocessed_client(cl):
                cl.connection = connection
                cl.catalogClient = CatalogClient(cl)
                cl.budgetClient = BudgetClient(cl)
                return cl

            previous_client = self.session.query(nYnabClient_).get(client_id)
            if previous_client is not None:
                previous_client.session = self.session
                return postprocessed_client(previous_client)

            client = nYnabClient_(id=client_id, budget_name=args.budget_name)
            client.engine = self.engine
            client.session = self.session
            client.add_missing()
            client = postprocessed_client(client)



            self.session.add(client)
            client.session.commit()
            if sync:
                client.sync()
            return client
        except BudgetNotFound:
            LOG.error('No budget by the name %s found in nYNAB' % args.budgetname)
            raise


def clientfromargs(args, sync=True):
    return nYnabClientFactory().create_client(args, sync)


from pynYNAB.schema.Client import nYnabClient_
from pynYNAB.schema.budget import Account
from pynYNAB.schema.roots import Budget
from tests.test import CommonTest


class TestGetChangedEntities(CommonTest):
    def setUp(self):
        super(TestGetChangedEntities, self).setUp()
        self.obj = Budget()
        self.account = Account()
        self.obj.be_accounts = [self.account]
        self.obj.clear_changed_entities()
        self.account2 = Account(id=self.account)

    def testgetChangedEntities_add(self):
        added_account = Account()
        self.obj.be_accounts.append(added_account)
        changed_entities = self.obj.get_changed_entities()
        self.assertEqual(changed_entities, {'be_accounts': [added_account]})

    def testgetChangedEntities_delete(self):
        self.obj.be_accounts.remove(self.account)
        changed_entities = self.obj.get_changed_entities()
        deleted = self.account.copy()
        deleted.is_tombstone = True
        self.assertEqual(changed_entities, {'be_accounts': [deleted]})

    def testgetChangedEntities_modify(self):
        self.account.account_name = 'BLA'
        changed_entities = self.obj.get_changed_entities()
        self.assertEqual(changed_entities, {'be_accounts': [self.account]})


class TestUpdateChangedEntities(CommonTest):
    def setUp(self):
        super(TestUpdateChangedEntities, self).setUp()
        self.account = Account()
        self.client = nYnabClient_(budgetname='Mock Budget')
        self.client.budget.be_accounts = [self.account]
        self.account2 = self.account.copy()
        self.client.session.commit()

    def testupdateChangedEntities_add(self):
        new_account = Account()
        changed_entities = dict(
            be_accounts=[new_account]
        )
        self.client.budgetClient.update_from_changed_entities(changed_entities)
        self.assertEqual(len(self.client.budget.be_accounts), 2)
        self.assertIn(new_account, self.client.budget.be_accounts)

    def testupdateChangedEntities_delete(self):
        self.account2.is_tombstone = True
        changed_entities = dict(
            be_accounts=[self.account2]
        )
        self.client.budgetClient.update_from_changed_entities(changed_entities)
        self.assertEqual(len(self.client.budget.be_accounts), 0)

    def testupdateChangedEntities_modify(self):
        self.account2 = self.account.copy()
        self.account2.note = 'note'
        changed_entities = dict(
            be_accounts=[self.account2]
        )

        self.client.budgetClient.update_from_changed_entities(changed_entities)
        self.assertEqual(len(self.client.budget.be_accounts), 1)
        acc = self.client.budget.be_accounts[0]
        self.assertEqual(acc, self.account2)





from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship

from pynYNAB.schema.Entity import Entity, AccountTypes, Base
from pynYNAB.schema.types import AmountType


class BudgetEntity(Entity):
    @declared_attr
    def parent_id(self):
        return Column(ForeignKey('budget.id'))

    @declared_attr
    def parent(self):
        return relationship('Budget', lazy=False)


class Transaction(Base, BudgetEntity):
    accepted = Column(Boolean, default=True)
    amount = Column(AmountType)
    cash_amount = Column(AmountType)
    check_number = Column(String)
    cleared = Column(String, default='Uncleared')
    credit_amount = Column(AmountType)
    credit_amount_adjusted = Column(AmountType)
    date = Column(Date)
    date_entered_from_schedule = Column(Date)
    entities_account_id = Column(ForeignKey('account.id'))
    entities_account = relationship('Account', foreign_keys=entities_account_id)
    entities_payee_id = Column(ForeignKey('payee.id'))
    entities_payee = relationship('Payee')
    entities_scheduled_transaction_id = Column(ForeignKey('scheduledtransaction.id'))
    entities_subcategory_id = Column(ForeignKey('subcategory.id'))
    entities_subcategory = relationship('SubCategory')
    flag = Column(String, default=None)
    imported_date = Column(Date)
    imported_payee = Column(String)
    matched_transaction_id = Column(ForeignKey('transaction.id'))
    matched_transaction = relationship('Transaction', foreign_keys=matched_transaction_id)
    memo = Column(String)
    source = Column(String)
    subcategory_credit_amount_preceding = Column(AmountType, default=0)
    transfer_account_id = Column(ForeignKey('account.id'))
    transfer_account = relationship('Account', foreign_keys=transfer_account_id)
    transfer_subtransaction_id = Column(ForeignKey('subtransaction.id'))
    transfer_transaction_id = Column(ForeignKey('transaction.id'))
    transfer_transaction = relationship('Transaction', foreign_keys=transfer_transaction_id)
    ynab_id = Column(String)


class MasterCategory(Base, BudgetEntity):
    deletable = Column(Boolean, default=True)
    internal_name = Column(String)
    is_hidden = Column(Boolean, default=False)
    name = Column(String)
    note = Column(String)
    sortable_index = Column(String)


class Setting(Base, BudgetEntity):
    setting_name = Column(String)
    setting_value = Column(String)


class MonthlyBudgetCalculation(Base, BudgetEntity):
    additional_to_be_budgeted = Column(AmountType)
    age_of_money = Column(String)
    available_to_budget = Column(String)
    balance = Column(String)
    budgeted = Column(String)
    calculation_notes = Column(String)
    cash_outflows = Column(AmountType)
    credit_outflows = Column(AmountType)
    deferred_income = Column(AmountType)
    entities_monthly_budget_id = Column(ForeignKey('monthlybudget.id'))
    entities_monthly_budget = relationship('MonthlyBudget')
    hidden_balance = Column(AmountType)
    hidden_budgeted = Column(AmountType)
    hidden_cash_outflows = Column(AmountType)
    hidden_credit_outflows = Column(AmountType)
    immediate_income = Column(AmountType)
    over_spent = Column(AmountType)
    previous_income = Column(AmountType)
    uncategorized_balance = Column(AmountType)
    uncategorized_cash_outflows = Column(AmountType)
    uncategorized_credit_outflows = Column(AmountType)


class AccountMapping(Base, BudgetEntity):
    date_sequence = Column(Date)
    entities_account_id = Column(ForeignKey('account.id'))
    entities_account = relationship('Account', foreign_keys=entities_account_id)
    hash = Column(String)
    fid = Column(String)
    salt = Column(String)
    shortened_account_id = Column(String)
    should_flip_payees_memos = Column(String)
    should_import_memos = Column(String)
    skip_import = Column(String)


class Subtransaction(Base, BudgetEntity):
    amount = Column(AmountType)
    cash_amount = Column(AmountType)
    check_number = Column(String)
    credit_amount = Column(AmountType)
    entities_payee_id = Column(ForeignKey('payee.id'))
    entities_payee = relationship('Payee')
    entities_subcategory_id = Column(ForeignKey('subcategory.id'))
    entities_subcategory = relationship('SubCategory', foreign_keys=entities_subcategory_id)
    entities_transaction_id = Column(ForeignKey('transaction.id'))
    entities_transaction = relationship('Transaction', foreign_keys=entities_transaction_id)
    memo = Column(String)
    sortable_index = Column(Integer, default=0)
    subcategory_credit_amount_preceding = Column(Integer, default=0)
    transfer_account_id = Column(ForeignKey('account.id'))
    transfer_account = relationship('Account')
    transfer_transaction_id = Column(ForeignKey('transaction.id'))
    transfer_transaction = relationship('Transaction', backref='transfer_subtransaction',
                                        foreign_keys=transfer_transaction_id)


class ScheduledSubtransaction(Base, BudgetEntity):
    amount = Column(AmountType)
    entities_payee_id = Column(ForeignKey('payee.id'))
    entities_payee = relationship('Payee', foreign_keys=entities_payee_id)
    entities_scheduled_transaction_id = Column(ForeignKey('scheduledtransaction.id'))
    entities_scheduled_transaction = relationship('ScheduledTransaction',
                                                  foreign_keys=entities_scheduled_transaction_id)
    entities_subcategory_id = Column(ForeignKey('subcategory.id'))
    entities_subcategory = relationship('SubCategory', foreign_keys=entities_subcategory_id)
    memo = Column(String)
    sortable_index = Column(Integer, default=0)
    transfer_account_id = Column(ForeignKey('account.id'))
    transfer_account = relationship('Account', foreign_keys=transfer_account_id)


class MonthlyBudget(Base, BudgetEntity):
    month = Column(Date)
    note = Column(String)


class SubCategory(Base, BudgetEntity):
    entities_account_id = Column(ForeignKey('account.id'))
    entities_account = relationship('Account', foreign_keys=entities_account_id)
    entities_master_category_id = Column(ForeignKey('mastercategory.id'))
    entities_master_category = relationship('MasterCategory', foreign_keys=entities_master_category_id)
    goal_creation_month = Column(String)
    goal_type = Column(String)
    internal_name = Column(String)
    is_hidden = Column(Boolean, default=False)
    monthly_funding = Column(String)
    name = Column(String)
    note = Column(String)
    sortable_index = Column(Integer, default=0)
    target_balance = Column(AmountType)
    target_balance_month = Column(String)
    type = Column(String)


class PayeeLocation(Base, BudgetEntity):
    entities_payee_id = Column(ForeignKey('payee.id'))
    entities_payee = relationship('Payee', foreign_keys=entities_payee_id)
    latitude = Column(String)
    longitude = Column(String)


class AccountCalculation(Base, BudgetEntity):
    cleared_balance = Column(AmountType)
    entities_account_id = Column(ForeignKey('account.id'))
    entities_account = relationship('Account', foreign_keys=entities_account_id)
    error_count = Column(String)
    info_count = Column(String)
    transaction_count = Column(String)
    uncleared_balance = Column(AmountType)
    warning_count = Column(String)


class MonthlyAccountCalculation(Base, BudgetEntity):
    cleared_balance = Column(AmountType)
    entities_account_id = Column(ForeignKey('account.id'))
    entities_account = relationship('Account', foreign_keys=entities_account_id)
    error_count = Column(String)
    info_count = Column(String)
    month = Column(String)
    transaction_count = Column(String)
    uncleared_balance = Column(AmountType)
    warning_count = Column(String)
    rolling_balance = Column(AmountType)


class MonthlySubcategoryBudgetCalculation(Base, BudgetEntity):
    additional_to_be_budgeted = Column(String)
    all_spending = Column(AmountType)
    all_spending_since_last_payment = Column(AmountType)
    balance = Column(AmountType)
    balance_previous_month = Column(AmountType)
    budgeted_average = Column(AmountType)
    budgeted_cash_outflows = Column(AmountType)
    budgeted_credit_outflows = Column(AmountType)
    budgeted_previous_month = Column(AmountType)
    budgeted_spending = Column(AmountType)
    cash_outflows = Column(AmountType)
    credit_outflows = Column(AmountType)
    entities_monthly_subcategory_budget_id = Column(ForeignKey('monthlysubcategorybudget.id'))
    entities_monthly_subcategory_budget = relationship('MonthlySubcategoryBudget',
                                                       foreign_keys=entities_monthly_subcategory_budget_id)
    goal_expected_completion = Column(String)
    goal_overall_funded = Column(AmountType)
    goal_overall_left = Column(AmountType)
    goal_percentage_complete = Column(String)
    goal_target = Column(String)
    goal_under_funded = Column(String)
    overspending_affects_buffer = Column(String)
    payment_average = Column(AmountType)
    payment_previous_month = Column(AmountType)
    spent_average = Column(AmountType)
    spent_previous_month = Column(AmountType)
    unbudgeted_cash_outflows = Column(AmountType)
    unbudgeted_credit_outflows = Column(AmountType)
    upcoming_transactions = Column(AmountType)
    upcoming_transactions_count = Column(String)
    positive_cash_outflows = Column(AmountType)


class UpComingInstance(Base):
    scheduledtransaction_id = Column(ForeignKey('scheduledtransaction.id'))


class ScheduledTransaction(Base, BudgetEntity):
    amount = Column(AmountType)
    date = Column(Date)
    entities_account_id = Column(ForeignKey('account.id'))
    entities_account = relationship('Account', foreign_keys=entities_account_id)
    entities_payee_id = Column(ForeignKey('payee.id'))
    entities_payee = relationship('Payee', foreign_keys=entities_payee_id)
    entities_subcategory_id = Column(ForeignKey('subcategory.id'))
    entities_subcategory = relationship('SubCategory', foreign_keys=entities_subcategory_id)
    transaction = relationship(Transaction, backref='entities_scheduled_transaction')
    flag = Column(String, default=None)
    frequency = Column(String)
    memo = Column(String)
    transfer_account_id = Column(ForeignKey('account.id'))
    transfer_account = relationship('Account', foreign_keys=transfer_account_id)

    upcoming_instances = relationship('UpComingInstance')


class Payee(Base, BudgetEntity):
    auto_fill_amount = Column(AmountType)
    auto_fill_amount_enabled = Column(String)
    auto_fill_memo = Column(String)
    auto_fill_memo_enabled = Column(String)
    auto_fill_subcategory_enabled = Column(String)
    auto_fill_subcategory_id = Column(ForeignKey('subcategory.id'))
    auto_fill_subcategory = relationship('SubCategory', foreign_keys=auto_fill_subcategory_id)
    enabled = Column(Boolean, default=True)
    entities_account_id = Column(ForeignKey('account.id'))
    entities_account = relationship('Account', foreign_keys=entities_account_id)
    internal_name = Column(String)
    name = Column(String)
    rename_on_import_enabled = Column(String)


class MonthlySubcategoryBudget(Base, BudgetEntity):
    budgeted = Column(AmountType)
    entities_monthly_budget_id = Column(ForeignKey('monthlybudget.id'))
    entities_monthly_budget = relationship('MonthlyBudget', foreign_keys=entities_monthly_budget_id)
    entities_subcategory_id = Column(ForeignKey('subcategory.id'))
    entities_subcategory = relationship('SubCategory', foreign_keys=entities_subcategory_id)
    note = Column(String)
    overspending_handling = Column(String)


class TransactionGroup(dict):
    def get_apidict(self):
        return self

    def get_dict(self):
        return self


class PayeeRenameCondition(Base, BudgetEntity):
    entities_payee_id = Column(ForeignKey('payee.id'))
    entities_payee = relationship('Payee', foreign_keys=entities_payee_id)
    operand = Column(String)
    operator = Column(String)


class Account(Base, BudgetEntity):
    account_name = Column(String)
    account_type = Column(Enum(AccountTypes), default=AccountTypes.undef)
    hidden = Column(Boolean, default=False)
    last_entered_check_number = Column(String)
    last_reconciled_balance = Column(String)
    last_reconciled_date = Column(Date)
    note = Column(String)
    sortable_index = Column(Integer, default=0)
    on_budget = Column(Boolean, default=True)

    direct_connect_enabled = Column(Boolean, default=False)
    direct_connect_account_id = Column(String)
    direct_connect_institution_id = Column(String)
    direct_connect_last_imported_at = Column(Date)
    direct_connect_last_error_code = Column(String)

    def get_dict(self):
        super_dict = super(Account, self).get_dict()
        if not super_dict['direct_connect_enabled']:
            super_dict['direct_connect_enabled'] = False
            del super_dict['direct_connect_account_id']
            del super_dict['direct_connect_last_error_code']
            del super_dict['direct_connect_institution_id']
            del super_dict['direct_connect_last_imported_at']
        return super_dict



import random

from datetime import datetime

from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.__main__ import parser
from pynYNAB.schema.budget import Transaction

# used to ping the nYNAB API to check that the sync works

args = parser.parse_known_args()[0]
N = 20
client = clientfromargs(args)
client.sync()
account = next(acc for acc in client.budget.be_accounts)
for _ in range(0, N):
    transaction = Transaction(
        cleared='Uncleared',
        date=datetime.now(),
        entities_account=account,
        amount=random.randrange(-10, 10)
    )
    client.add_transaction(transaction)



import unittest
import uuid
import re
from datetime import datetime

from pynYNAB.schema.Entity import date_from_api


strdate0 = '2015-04-03'
date0 = datetime.strptime(strdate0, '%Y-%m-%d').date()
datestrings = [strdate0+'T00:00:00',strdate0,'random+'+strdate0]


class TestDatesCases(unittest.TestCase):
    def do_test(self, x, test_name):
        self.assertEqual(date_from_api(None, x), date0, msg=test_name)

    def test_dates(self):
        pass
        # Test some other functionality here

        for s in datestrings:
            test_name = 'test_datefromapi_%s' % re.sub('[^a-zA-Z0-9]+', '', s)
            self.do_test(s, test_name)



from datetime import datetime
from ofxtools import OFXTree

from pynYNAB.ClientFactory import clientfromargs
from pynYNAB.schema.budget import Transaction, Payee


def do_ofximport(args, client=None):
    delta = 0
    if client is None:
        client = clientfromargs(args)

    tree = OFXTree()
    tree.parse(args.ofxfile)
    response = tree.convert()
    stmts = response.statements

    accounts = client.budget.be_accounts
    accountvsnotes = {account.note: account for account in accounts if account.note is not None}

    for stmt in stmts:
        key = stmt.account.bankid + ' ' + stmt.account.branchid + ' ' + stmt.account.acctid
        if all(key not in note for note in accountvsnotes):
            if len(accounts) == 0:
                print('No accounts available in this budget')
                exit(-1)

            # ask user input for which bank account this is, then save it into the account note in nYNAB
            account = client.select_account_ui()

            # Save the selection in the nYNAB account note
            addon = 'key[' + key + ']key'
            if account.note is not None:
                account.note += addon
            else:
                account.note = addon
        else:
            for note in accountvsnotes:
                if key in note:
                    account = accountvsnotes[note]

                    imported_date = datetime.now().date()

                    for ofx_transaction in stmt.transactions:
                        payee_name = ofx_transaction.name if ofx_transaction.payee is None else ofx_transaction.payee
                        try:
                            payee = next(p for p in client.budget.be_payees if p.name == payee_name)
                        except StopIteration:
                            payee = Payee(
                                name=payee_name
                            )
                            client.budget.be_payees.append(payee)
                            delta += 1

                        # use ftid so we don't import duplicates
                        if not any(ofx_transaction.fitid in transaction.memo for transaction in
                                   client.budget.be_transactions if
                                   transaction.memo is not None and transaction.entities_account_id == account.id):
                            transaction = Transaction(
                                date=ofx_transaction.dtposted,
                                memo=ofx_transaction.memo + '    ' + ofx_transaction.fitid,
                                imported_payee=payee_name,
                                entities_payee_id=payee.id,
                                imported_date=imported_date,
                                source="Imported",
                                check_number=ofx_transaction.checknum,
                                amount=float(ofx_transaction.trnamt),
                                entities_account_id=account.id
                            )
                            client.budget.be_transactions.append(transaction)
                            delta += 1

    return delta


if __name__ == "__main__":
    from pynYNAB.__main__ import MainCommands
    MainCommands.ofximport()
